#include "Defs.h"
#include "Ray.h"
#include "Point.h"
#include "Face.h"
#include "Edge.h"
#include "RGB.h"
#include "Hood.h"

#ifndef DATASET_H_INCLUDED
#define DATASET_H_INCLUDED

class Renderer;

class Dataset
{
public:
  int nx, ny, nz;
  float sx, sy, sz;
  float maxx, maxy, maxz;
  int imaxx, imaxy, imaxz;
  int minv, maxv;
  int dbg;
  Renderer *R;
  Face *Faces[6];
  Edge *Edges[12];
  Point *Vertices[8];
  unsigned char ***Data;
  unsigned char *Mask;
  Point ***Deformation;
  unsigned char *DeformationMask;
  Point avp, MinDef, MaxDef;
  int avcx, avcy, avcz;
  int mindefx, mindefy, mindefz, maxdefx, maxdefy, maxdefz;

  Dataset(Renderer *r);
  ~Dataset();
  void loadData();
  void createMask();
  int intersect(Ray *r);
  int getExtremeIntersections(Ray *ray, float &first, float &last);
  float getVoxelValue(int x, int y, int z);
  float getValue(Point *p);
  int isInside(Point *p);
  int onFace(Point *p);
  int isOnEdge(Point *p);
  void computeDeformation();
  void distributeDeformedValue(Point *p, unsigned char v);
  int moveVoxel(int xi, int yi, int zi, int level);
  int isDeformed(int x, int y, int z);
  void setDeformed(int x, int y, int z);
  void getPosition(int x, int y, int z, Point *p);
  int goodDeformedVoxel(Point *v, Point *p);

  int isValueChanged(int x, int y, int z);
  void setValueChanged(int x, int y, int z);

  int getHood(Point *p, Hood *h);
  void getVoxelNormal(int x, int y, int z, Point *n);
  int getPointNormal(Point *p, Point *n);
  inline unsigned char getMax(unsigned char a1, unsigned char a2, 
                              unsigned char a3, unsigned char a4, 
                              unsigned char a5, unsigned char a6, 
                              unsigned char a7, unsigned char a8);

  inline unsigned char getMin(unsigned char a1, unsigned char a2, 
                              unsigned char a3, unsigned char a4, 
                              unsigned char a5, unsigned char a6, 
                              unsigned char a7, unsigned char a8);
};

#endif
