#include "Face.h"

/* ------------------------------------------------------------------------ */
Face::Face(Point *p0, Point *p1, Point *p2, Point *p3)
{
  int i;

  Points[0] = new Point(p0);
  Points[1] = new Point(p1);
  Points[2] = new Point(p2);
  Points[3] = new Point(p3);

  minx = 100000.0f;
  miny = 100000.0f;
  minz = 100000.0f;
  maxx = 0.0f;
  maxy = 0.0f;
  maxz = 0.0f;

  for(i=0; i<4; i++)
  {
    if(minx > Points[i]->X)
      minx = Points[i]->X;
    if(maxx < Points[i]->X)
      maxx = Points[i]->X;
    if(miny > Points[i]->Y)
      miny = Points[i]->Y;
    if(maxy < Points[i]->Y)
      maxy = Points[i]->Y;
    if(minz > Points[i]->Z)
      minz = Points[i]->Z;
    if(maxz < Points[i]->Z)
      maxz = Points[i]->Z;
  }

  /* Equation of the plane A*x+B*y+C*z+D=0 */
  A = p0->Y*(p1->Z - p2->Z) + p1->Y*(p2->Z - p0->Z) + p2->Y*(p0->Z - p1->Z);
  B = p0->Z*(p1->X - p2->X) + p1->Z*(p2->X - p0->X) + p2->Z*(p0->X - p1->X);
  C = p0->X*(p1->Y - p2->Y) + p1->X*(p2->Y - p0->Y) + p2->X*(p0->Y - p1->Y);
  D = - p0->X*(p1->Y*p2->Z - p2->Y*p1->Z) - 
    p1->X*(p2->Y*p0->Z - p0->Y*p2->Z) - 
    p2->X*(p0->Y*p1->Z - p1->Y*p0->Z);
}

/* ------------------------------------------------------------------------ */
int Face::getIntersection(Ray *ray, float &t)
{
  float a, b, xi, yi, zi, trhl;

  /* Intersetion equation between the plane and the ray a*t+b=0 */
  a = A*ray->Dx + B*ray->Dy + C*ray->Dz;
  b = A*ray->X0 + B*ray->Y0 + C*ray->Z0 + D;

  trhl = 0.0001f;

  if(a != 0.0 && -b/a >= 0.0)
  {
    xi = ray->X0-b/a*ray->Dx;
    yi = ray->Y0-b/a*ray->Dy;
    zi = ray->Z0-b/a*ray->Dz;

    if(xi >= minx - trhl && xi <= maxx+trhl && 
       yi >= miny - trhl && yi <= maxy+trhl && 
       zi >= minz - trhl && zi <= maxz+trhl)
    {
      t = -b/a;
      return 1;
    }
  } 
  return 0;
}

/* ------------------------------------------------------------------------ */
int Face::near(Point *p, float dist)
{
  float faux;

  faux = (float)(A*p->X + B*p->Y + C*p->Z + D)/sqrt(A*A + B*B + C*C);
  
  if(fabs(faux) <= fabs(dist))
    return 1;
  return 0;
}

/* ------------------------------------------------------------------------ */
float Face::dist(Point *p)
{
  float faux;

  faux = (float)(A*p->X + B*p->Y + C*p->Z + D)/sqrt(A*A + B*B + C*C);
  
  return fabs(faux);
}
