#include "Point.h"
#include "Transform.h"

/* ------------------------------------------------------------------------ */
Point::Point(Point *p)
{
  if(p != NULL)
  {
    X = p->X;
    Y = p->Y;
    Z = p->Z;
  }
  else
    Point();
}

/* ------------------------------------------------------------------------ */
void Point::copy(Point *p)
{
  if(p != NULL)
  {
    X = p->X;
    Y = p->Y;
    Z = p->Z;
  }
  else
  {
    X = 0.0f;
    Y = 0.0f;
    Z = 0.0f;
  }
}

/* ------------------------------------------------------------------------ */
Point::Point(float x, float y, float z)
{
  X = x;
  Y = y;
  Z = z;
}

/* ------------------------------------------------------------------------ */
void Point::copy(float x, float y, float z)
{
  X = x;
  Y = y;
  Z = z;
}

/* ------------------------------------------------------------------------ */
void Point::init()
{
  X = 0.0f;
  Y = 0.0f;
  Z = 0.0f;
}

/* ------------------------------------------------------------------------ */
void Point::normalize()
{
  float m;
  
  m = (float)sqrt(X*X+Y*Y+Z*Z);
  if(m < 0.000001f)
    return;
  X = X/m;
  Y = Y/m;
  Z = Z/m;
}

/* ------------------------------------------------------------------------ */
float Point::dot(Point *p)
{
  return (float)fabs(X*p->X + Y*p->Y + Z*p->Z);
}

/* ------------------------------------------------------------------------ */
void Point::cross(Point *p)
{
  float x, y, z;

  x = Y*p->Z - Z*p->Y;
  y = Z*p->X - X*p->Z;
  z = X*p->Y - Y*p->X;

  X = x;
  Y = y;
  Z = z;
}

/* ------------------------------------------------------------------------ */
void Point::add(Point *p)
{
  X += p->X;
  Y += p->Y;
  Z += p->Z;
}

/* ------------------------------------------------------------------------ */
void Point::sub(Point *p)
{
  X -= p->X;
  Y -= p->Y;
  Z -= p->Z;
}

/* ------------------------------------------------------------------------ */
float Point::dist(Point *p)
{
  return (float)sqrt((X-p->X)*(X-p->X) + (Y-p->Y)*(Y-p->Y) + (Z-p->Z)*(Z-p->Z));
}

/* ------------------------------------------------------------------------ */
float Point::dist(float x, float y, float z)
{
  return (float)sqrt((X-x)*(X-x) + (Y-y)*(Y-y) + (Z-z)*(Z-z));
}

/* ------------------------------------------------------------------------ */
int Point::parallel(Point *p)
{
  Point p1, p2;

  p1.copy(p);
  p1.normalize();
  p2.copy(this);
  p2.normalize();

  if(p1.dot(&p2) >= 1)
    return 1;
  return 0;
}

/* ------------------------------------------------------------------------ */
void Point::scale(float s)
{
  X *= s;
  Y *= s;
  Z *= s;
}
/* ------------------------------------------------------------------------ */
void Point::rot(Rotation *r)
{
  float x, y, z;
  
  x = r->m[0][0]*X + r->m[0][1]*Y + r->m[0][2]*Z;
  y = r->m[1][0]*X + r->m[1][1]*Y + r->m[1][2]*Z;
  z = r->m[2][0]*X + r->m[2][1]*Y + r->m[2][2]*Z;

  X = x;
  Y = y;
  Z = z;
}

/* ------------------------------------------------------------------------ */
void Point::trans(Transform *t)
{
  float x, y, z;
  
  x = t->m[0][0]*X + t->m[0][1]*Y + t->m[0][2]*Z + t->m[0][3];
  y = t->m[1][0]*X + t->m[1][1]*Y + t->m[1][2]*Z + t->m[1][3];
  z = t->m[2][0]*X + t->m[2][1]*Y + t->m[2][2]*Z + t->m[2][3];

  X = x;
  Y = y;
  Z = z;
}

/* ------------------------------------------------------------------------ */
float Point::magnitude()
{
  return (float)sqrt(X*X+Y*Y+Z*Z);
}

/* ------------------------------------------------------------------------ */
int Point::greaterThan(Point *p)
{
  if(X >= p->X && Y >= p->Y && Z >= p->Z)
    return 1;
  return 0;
}

/* ------------------------------------------------------------------------ */
int Point::lowerThan(Point *p)
{
  if(X <= p->X && Y <= p->Y && Z <= p->Z)
    return 1;
  return 0;
}

/* ------------------------------------------------------------------------ */
void Point::print()
{
  printf("%f   %f   %f\n", X, Y, Z);
}
