#include "Renderer.h"

/* ------------------------------------------------------------------------ */
Renderer::Renderer(char *dn)
{
  Point pi, pj;
  FILE *fi;
  char saux[100];
  RGBItem *rgbi;
  AlphaItem *ai;
  int i, j, x, y;
  float d;
  Point px, py, pz;
  Rotation r;

  sprintf(saux, "../data/%s.conf", dn);
  fi = fopen(saux, "rt");
  if(fi == NULL)
  {
    printf("Cannot find the dataset configuration file. Abort.\n");
    exit(0);
  }
  fgets(DatasetName, 199, fi);
  DatasetName[strlen(DatasetName)-1] = 0;
  fgets(RGBFile, 199, fi);
  RGBFile[strlen(RGBFile)-1] = 0;
  fgets(AlphaFile, 199, fi);
  AlphaFile[strlen(AlphaFile)-1] = 0;
  fgets(StfFile, 199, fi);
  StfFile[strlen(StfFile)-1] = 0;
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &minx);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &maxx);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &miny);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &maxy);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &minz);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &maxz);
  fgets(saux, 99, fi);
  sscanf(saux, "%f", &Stride);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &WSize);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &EUS);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &HaveShading);
  fgets(saux, 99, fi);
  sscanf(saux, "%d", &HaveDeformation);
  fgets(saux, 99, fi);
  sscanf(saux, "%f %f %f %f", &KA, &KD, &KS, &SPECN);
  fgets(saux, 99, fi);
  sscanf(saux, "%f", &HIOAlpha);
  fgets(saux, 99, fi);
  sscanf(saux, "%f", &HIO.Radius);
  fgets(saux, 99, fi);
  sscanf(saux, "%f %f %f", &HIOColor.Red, &HIOColor.Green, &HIOColor.Blue);
  fgets(saux, 99, fi);
  sscanf(saux, "%f %f %f", &DIR.X, &DIR.Y, &DIR.Z);
  fgets(saux, 99, fi);
  sscanf(saux, "%f %f %f", &LPOS.X, &LPOS.Y, &LPOS.Z);
  fclose(fi);
  Background = new RGB(0.0f, 0.6f, 0.9f);

  fi = fopen(RGBFile, "rt");
  while(!feof(fi))
  {
    rgbi = new RGBItem();
    if(!fgets(saux, 99, fi))
      break;
    sscanf(saux, "%f %f %f %f %f", &rgbi->min, &rgbi->max, 
           &rgbi->rgb->Red, &rgbi->rgb->Green, &rgbi->rgb->Blue);
    if(rgbi->rgb->Red < 0.0f || rgbi->rgb->Green < 0.0f || 
       rgbi->rgb->Blue < 0.0f)
    {
      delete rgbi->rgb;
      rgbi->rgb = NULL;
    }
    RGBMap.addToTail(rgbi);
  }
  fclose(fi);

  fi = fopen(AlphaFile, "rt");
  while(!feof(fi))
  {
    ai = new AlphaItem();
    if(!fgets(saux, 99, fi))
      break;
    sscanf(saux, "%f %f %f", &ai->min, &ai->max, &ai->alpha);
    AlphaMap.addToTail(ai);
  }
  fclose(fi);

  DS = new Dataset(this);

  if(minx < 0)
    minx = 0;
  if(miny < 0)
    miny = 0;
  if(minz < 0)
    minz = 0;

  if(maxx > DS->nx-1)
    maxx = DS->nx-1;
  if(maxy > DS->ny-1)
    maxy = DS->ny-1;
  if(maxz > DS->nz-1)
    maxz = DS->nz-1;

  DSize = 0.0f;
  for(i=0; i<8; i++)
  {
    pi.copy(DS->Vertices[i]);
    for(j=0; j<8; j++)
    {
      pj.copy(DS->Vertices[j]);
      pj.sub(&pi);
      d = pj.magnitude();
      if(DSize < d)
        DSize = d;
    }
  }

  DSize *= 1.1f;

  INITDIR.copy(&DIR);
  CENTER.copy((maxx-minx)/2.0f*DS->sx, 
                 (maxy-miny)/2.0f*DS->sy,
                 (maxz-minz)/2.0f*DS->sz);

  VP.copy((maxx-minx)/2.0f*DS->sx, 
                 (maxy-miny)/2.0f*DS->sy,
                 (maxz-minz)/2.0f*DS->sz);
  VP.add(&DIR);

  Dist = DIR.magnitude();
  DIR.normalize();
  DIR.scale(-1.0f);

  VUP.copy(&DIR);
  VUP.Y -= 100;

  if(DIR.parallel(&VUP))
  {
    VUP.X = DIR.X;
    VUP.Y = DIR.Y;
    VUP.Z = DIR.Z+100;
  }
  VUP.normalize();


  pz.copy(&DIR);

  px.copy(&DIR);
  px.cross(&VUP);
  px.normalize();

  py.copy(&DIR);
  py.cross(&px);
  py.normalize();

  r.m[0][0] = px.X;
  r.m[1][0] = px.Y;
  r.m[2][0] = px.Z;

  r.m[0][1] = py.X;
  r.m[1][1] = py.Y;
  r.m[2][1] = py.Z;

  r.m[0][2] = pz.X;
  r.m[1][2] = pz.Y;
  r.m[2][2] = pz.Z;

  T.copy(&r, &VP);

  EUMaxRGB = new float[EUS];
  PrevMaxRGB = 1.0f;

  Image = new (RGB*)[WSize];
  ImageMask = new (unsigned char*)[WSize];
  for(y=0; y<WSize; y++)
  {
    Image[y] = new RGB[WSize];
    ImageMask[y] = new unsigned char[WSize];
    for(x=0; x<WSize; x++)
    {
      ImageMask[y][x] = 1;
      Image[y][x] = new RGB();
    }
  }

  FrameCount = 0;
  AverageFrameDuration = 0.0;
}

/* ------------------------------------------------------------------------ */
Renderer::~Renderer()
{
  delete DS;
}

/* ------------------------------------------------------------------------ */
void Renderer::setLightPosition(Point *lpos)
{
  LPOS.copy(lpos);
}

/* ------------------------------------------------------------------------ */
void Renderer::setViewPlane(Transform *t)
{
  T.copy(t);
}

/* ------------------------------------------------------------------------ */
void Renderer::render()
{
  timeval begin, end;
  double duration;

  gettimeofday(&begin, NULL);
  renderRegion(0, 0, WSize-1, WSize-1);
  gettimeofday(&end, NULL);
  duration = end.tv_sec*1000*1000 + end.tv_usec - 
    begin.tv_sec*1000*1000 - begin.tv_usec;
  duration /= 1000;
  FrameCount++;
  AverageFrameDuration = (FrameCount-1)/FrameCount*AverageFrameDuration+
    duration/FrameCount;
}

/* ------------------------------------------------------------------------ */
void Renderer::renderHIO()
{
  Transform tr;
  Point p;
  float minx, miny, maxx, maxy;
  
  tr.copy(&T);
  tr.invert();

  p.copy(&HIO.Center);
  p.sub(&CENTER);
  p.trans(&tr);

  p.X = (float)WSize/DSize*p.X + (float)WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + (float)WSize/2.0f;

  minx = p.X;
  miny = p.Y;
  maxx = p.X;
  maxy = p.Y;

  if(minx > PrevWindowHIP.X)
    minx =  PrevWindowHIP.X;
  if(miny > PrevWindowHIP.Y)
    miny =  PrevWindowHIP.Y;
  if(maxx < PrevWindowHIP.X)
    maxx =  PrevWindowHIP.X;
  if(maxy < PrevWindowHIP.Y)
    maxy =  PrevWindowHIP.Y;

  PrevWindowHIP.copy(&p);

  renderRegion((int)floor(minx-HIO.Radius) - 5, 
               (int)floor(miny-HIO.Radius) - 5,
               (int)ceil(maxx+HIO.Radius) + 5, 
               (int)ceil(maxy+HIO.Radius) + 5);
}

/* ------------------------------------------------------------------------ */
void Renderer::renderBox(Box *box)
{
  Transform tr;
  Point p;
  float minx, miny, maxx, maxy;
  
  tr.copy(&T);
  tr.invert();

  p.copy(box->minx, box->miny, box->minz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  minx = p.X;
  miny = p.Y;
  maxx = p.X;
  maxy = p.Y;

  p.copy(box->minx, box->miny, box->maxz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  if(p.X < minx)
    minx = p.X;
  if(p.Y < miny)
    miny = p.Y;
  if(p.X > maxx)
    maxx = p.X;
  if(p.Y > maxy)
    maxy = p.Y;

  p.copy(box->minx, box->maxy, box->minz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  if(p.X < minx)
    minx = p.X;
  if(p.Y < miny)
    miny = p.Y;
  if(p.X > maxx)
    maxx = p.X;
  if(p.Y > maxy)
    maxy = p.Y;

  p.copy(box->minx, box->maxy, box->maxz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  if(p.X < minx)
    minx = p.X;
  if(p.Y < miny)
    miny = p.Y;
  if(p.X > maxx)
    maxx = p.X;
  if(p.Y > maxy)
    maxy = p.Y;

  p.copy(box->maxx, box->miny, box->minz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  if(p.X < minx)
    minx = p.X;
  if(p.Y < miny)
    miny = p.Y;
  if(p.X > maxx)
    maxx = p.X;
  if(p.Y > maxy)
    maxy = p.Y;

  p.copy(box->maxx, box->miny, box->maxz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  if(p.X < minx)
    minx = p.X;
  if(p.Y < miny)
    miny = p.Y;
  if(p.X > maxx)
    maxx = p.X;
  if(p.Y > maxy)
    maxy = p.Y;

  p.copy(box->maxx, box->maxy, box->minz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  if(p.X < minx)
    minx = p.X;
  if(p.Y < miny)
    miny = p.Y;
  if(p.X > maxx)
    maxx = p.X;
  if(p.Y > maxy)
    maxy = p.Y;

  p.copy(box->maxx, box->maxy, box->maxz);
  p.sub(&CENTER);
  p.trans(&tr);
  p.X = (float)WSize/DSize*p.X + WSize/2.0f;
  p.Y = (float)WSize/DSize*p.Y + WSize/2.0f;
  if(p.X < minx)
    minx = p.X;
  if(p.Y < miny)
    miny = p.Y;
  if(p.X > maxx)
    maxx = p.X;
  if(p.Y > maxy)
    maxy = p.Y;

  renderRegion((int)floor(minx), (int)floor(miny),
               (int)ceil(maxx), (int)ceil(maxy));
}

/* ------------------------------------------------------------------------ */
void Renderer::renderRegion(int x0, int y0, int x1, int y1)
{
  int x, y, i;
  ExecUnitArgs *eua[EUS];
  pthread_t eus[EUS];
  pthread_attr_t attr;
  float max;

  if(x1 < 0 || y1 < 0 || x0 > WSize-1 || y0 > WSize-1 ||
     y1 < y0 || x1 < x0)
    return;

  if(x0 < 0)
    x0 = 0;
  if(y0 < 0)
    y0 = 0;
  if(x1 > WSize-1)
    x1 = WSize -1;
  if(y1 > WSize-1)
    y1 = WSize -1;

#ifdef ECHO_FEEDBACK
  timeval tvb, tve;

  printf("Rendering ... ");
  fflush(NULL);

  gettimeofday(&tvb, NULL);
#endif


  if(HaveDeformation)
    DS->computeDeformation();
  for(i=0; i<EUS; i++)
    EUMaxRGB[i] = 0.0f;

  for(y=y0; y<=y1; y++)
    for(x=x0; x<=x1; x++)
      ImageMask[y][x] = 1;
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, VHSCOPE);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  pthread_attr_setschedpolicy(&attr, VHSCHED);
  for(i=0; i<EUS; i++)
  {
    eua[i] = new ExecUnitArgs();
    eua[i]->renderer = this;
    eua[i]->offset = i;
    eua[i]->x0 = x0;
    eua[i]->x1 = x1+1;
    eua[i]->y0 = y0;
    eua[i]->y1 = y1+1;
    pthread_create(&eus[i], &attr, 
                   (void *(*)(void *))computePixelColor, eua[i]);
  }

  for(i=0; i<EUS; i++)
    pthread_join(eus[i], NULL);

  max = 0;
  for(i=0; i<EUS; i++)
    if(max < EUMaxRGB[i])
      max = EUMaxRGB[i];

  if(max < 1.0f)
    max = 1.0f;
  
  if(max < PrevMaxRGB)
    max = PrevMaxRGB;
  else
    PrevMaxRGB = max;

  for(y=0; y<WSize; y++)
    for(x=0; x<WSize; x++)
      if(ImageMask[x][y])
      {
        if(x0 <= x && y0 <= y && x1 >= x && y1 >= y)
        {
          Image[x][y].Red /=max;
          Image[x][y].Green /= max;
          Image[x][y].Blue /= max;
        }
        else
        {
          Image[x][y].Red *= PrevMaxRGB/max;
          Image[x][y].Green *= PrevMaxRGB/max;
          Image[x][y].Blue *= PrevMaxRGB/max;
        }
      }
  
#ifdef ECHO_FEEDBACK
  gettimeofday(&tve, NULL);
#endif

  pthread_attr_destroy(&attr);

#ifdef ECHO_FEEDBACK
  printf("Done.\n");

  printf("Duration: %d seconds and %d microseconds\n", 
         (int)(tve.tv_sec - tvb.tv_sec), (int)(tve.tv_usec - tvb.tv_usec));
  fflush(NULL);
#endif
}

/* ------------------------------------------------------------------------ */
void computePixelColor(ExecUnitArgs *eua)
{
  int wx, wy, x0, y0, x1, y1;
  Ray ray;
  RGB rgb, rgbaux, rv, rs;
  Point p, mp, V, N, L, H, DIR, NHIO;
  float NL, NH, first, last, t, alpha, v, transparency, faux, dist_hip, av, as;
  int pixelno, ofs, iaux;
  Renderer *R;
  List ray_intervals;
  Interval *intv_vol, *intv_hio, *intvaux;

  R = eua->renderer;

  DIR.copy(R->T.m[0][3], R->T.m[1][3], R->T.m[2][3]);
  DIR.sub(&R->CENTER);
  DIR.scale(-1);
  DIR.normalize();

  V.copy(R->T.m[0][3], R->T.m[1][3], R->T.m[2][3]);
  V.scale(-1);
  V.normalize();

  L.copy(&R->LPOS);
  L.normalize();

  H.copy(&V);
  H.add(&L);
  H.normalize();
  
  ofs = eua->offset;
  x0 = eua->x0;
  x1 = eua->x1;
  y0 = eua->y0;
  y1 = eua->y1;
  pixelno = 0;
  while((iaux = pixelno*R->EUS+ofs) < (y1-y0)*(x1-x0))
  {
    wy = iaux / (x1 - x0) + y0;
    wx = iaux % (x1 - x0) + x0;
    pixelno++;

    p.copy((wx - R->WSize/2.0f)*R->DSize/R->WSize, 
           (wy - R->WSize/2.0f)*R->DSize/R->WSize, 0);
    p.trans(&R->T);
    ray.copy(&p, &DIR, 1);

    /* Find the sampling intervals along the ray */
    if(R->DS->getExtremeIntersections(&ray, first, last))
    {
      intv_vol = new Interval(first, last);
      ray_intervals.addToHead(intv_vol);

      if(R->HIO.getExtremeIntersections(&ray, first, last))
      {
        intv_hio = new Interval(first, last);
        if(!intv_vol->merge(intv_hio))
        {
          if(intv_hio->xf < intv_vol->xi)
            ray_intervals.addToHead(intv_hio);
          else
            ray_intervals.addToTail(intv_hio);
        }
      }
    }
    else
    {
      if(R->HIO.getExtremeIntersections(&ray, first, last))
      {
        intv_hio = new Interval(first, last);
        ray_intervals.addToHead(intv_hio);
      }
    }
    
    if(ray_intervals.Head != NULL)
    {
      intvaux = (Interval*)ray_intervals.Head;
      transparency = 1.0f;
      alpha = 0.0f;
      rgb.copy(0.0f, 0.0f, 0.0f);
      rgbaux.copy(0.0f, 0.0f, 0.0f);

      while(intvaux != NULL && transparency > 0.0f)
      {
        first = intvaux->xi;
        last = intvaux->xf;
        for(t=first; t<last && transparency > 0.0f; t+=R->Stride)
        {
          ray.getPoint(t, &mp);

          as = 0.0f;
          av = 0.0f;
          alpha = 1.0f;
          rv.copy(0.0f, 0.0f, 0.0f);
          rs.copy(0.0f, 0.0f, 0.0f);

          dist_hip = R->HIO.Center.dist(&mp);
          if(dist_hip <= R->HIO.Radius)
          {
            as = R->HIOAlpha;
            rs.copy(&R->HIOColor);
          }

          if(R->DS->isInside(&mp))
          {
            v = R->DS->getValue(&mp);
            av = R->getAlpha(v);
            R->getRGB(v, &rv);
          }

          if(as+av > 1.0f)
            alpha = 1.0f;
          else
            alpha = as+av;
          
          if(alpha < 0.00001f)
            continue;
          rs.scale(as/(as+av));
          rv.scale(av/(as+av));
          rs.add(&rv);
          rgbaux.copy(&rs);

          if(R->HaveShading)
          {
            N.init();
            NHIO.init();
            if(dist_hip <= R->HIO.Radius)
            {
              R->HIO.getPointNormal(&mp, &NHIO);
              R->DS->getPointNormal(&mp, &N);
              N.add(&NHIO);
              N.normalize();
            }
            else
              if(!R->DS->getPointNormal(&mp, &N))
                continue;
          
            NL = N.dot(&L);
            NH = H.dot(&N);
          
            rgbaux.scale(R->KA + R->KD*NL + R->KS*pow(NH, R->SPECN));
          }
          rgbaux.scale(alpha*transparency);
          rgb.add(&rgbaux);
          transparency *= 1-alpha;
        }
        intvaux = (Interval*)intvaux->Next;
      }
      ray_intervals.clear();

      if(transparency != 0.0f)
      {
        rgbaux.copy(0.0f, 0.0f, 0.0f);
        rgbaux.scale(transparency);
        rgb.add(&rgbaux);
      }
        
      faux = R->getMax(rgb.Red, rgb.Green, rgb.Blue);
      if(R->EUMaxRGB[ofs] < faux)
        R->EUMaxRGB[ofs] = faux;

      R->Image[wx][wy].copy(&rgb);
    }
    else
    {
      R->Image[wx][wy].copy(&R->Background);
      R->ImageMask[wx][wy] = 0;
    }
  }
}

/* ------------------------------------------------------------------------ */
void Renderer::setHIPPosition(Point*p)
{
  if(p == NULL)
    return;

  PrevHIP.copy(&HIO.Center);
  HIO.Center.copy(p);
  DS->avp.copy(p);
}

/* ------------------------------------------------------------------------ */
void Renderer::setHIPPosition(float x, float y, float z)
{
  PrevHIP.copy(&HIO.Center);
  HIO.Center.copy(x, y, z);
//  DS->avp.copy(x, y, z);
}

/* ------------------------------------------------------------------------ */
void Renderer::setHIPVoxel(int x, int y, int z)
{
  DS->avcx = x;
  DS->avcy = y;
  DS->avcz = z;
}

/* ------------------------------------------------------------------------ */
void Renderer::getRGB(float v, RGB* rgb)
{
  RGBItem *raux;
  
  raux = (RGBItem *)RGBMap.Head;
  while(raux != NULL)
  {
    if(raux->min < v && raux->max >= v)
    {
      if(raux->rgb != NULL)
        rgb->copy(raux->rgb);
      else
      {
        if(v < 0.0f)
        {
          rgb->copy(0.0f, 1.0f, 0.0f);
          return;
        }
        if(v > 1.0f)
        {
          rgb->copy(1.0f, 0.0f, 0.0f);
          return;
        }
        rgb->copy(v, 1-v, 0.0f);
        return;
      }
    }
    raux = (RGBItem*) raux->Next;
  }
}

/* ------------------------------------------------------------------------ */
float Renderer::getAlpha(float v)
{
  AlphaItem *raux;

  if(v <= 0.00001f)
    return 0.0f;
  
  raux = (AlphaItem *)AlphaMap.Head;
  while(raux != NULL)
  {
    if(raux->min < v && v <= raux->max)
    {
      if(raux->alpha < 0.0f)
          return v;
      else
        return raux->alpha;
    }
    raux = (AlphaItem*) raux->Next;
  }
  return 0.0f;
}

/* ------------------------------------------------------------------------ */
void Renderer::writeImageFile(char s[])
{
  unsigned char r, g, b;
  int x, y;

  printf("Writing the file ... ");
  fflush(NULL);

  /* Open the image file */
  ImageFile = fopen(s, "wb");

  /* Write the PPM file header */
  fprintf(ImageFile, "P6\n");
  fprintf(ImageFile, "%d %d\n", WSize, WSize);
  fprintf(ImageFile, "255\n");

  for(y=0; y<WSize; y++)
    for(x=0; x<WSize; x++)
    {
      r = (unsigned char) floor(Image[x][y].Red*255);
      g = (unsigned char) floor(Image[x][y].Green*255);
      b = (unsigned char) floor(Image[x][y].Blue*255);

      fwrite(&r, 1, 1, ImageFile);
      fwrite(&g, 1, 1, ImageFile);
      fwrite(&b, 1, 1, ImageFile);
    }
  fclose(ImageFile);
  printf("Done.\n\n");
  fflush(NULL);
}

/* ------------------------------------------------------------------------ */
inline float Renderer::getMin(float a1, float a2, float a3)
{
  float m;

  if(a1 > 0.0f)
    m = a1;
  else
    m = 0.0f;
  if(m > a2)
    m = a2;
  if(m > a3)
    m = a3;
  return m;
}

/* ------------------------------------------------------------------------ */
inline float Renderer::getMax(float a1, float a2, float a3)
{
  float m;

  m = a1;
  if(m < a2)
    m = a2;
  if(m < a3)
    m = a3;
  return m;
}









