#include "Defs.h"
#include "RGB.h"
#include "Point.h"
#include "Ray.h"
#include "Dataset.h"
#include "Rotation.h"
#include "Transform.h"
#include "List.h"
#include "Box.h"
#include "Interval.h"
#include "Sphere.h"

#ifndef RENDERER_H_INCLUDED
#define RENDERER_H_INCLUDED

#define MINRGB -54321.0f
#define MAXRGB 54321.0f

class Renderer
{
public:
  Point VP, DIR, VUP, LPOS, INITDIR, CENTER;
  float KA, KD, KS, SPECN;
  Transform T;
  RGB Background;
  Dataset *DS;
  FILE *ImageFile;
  float DSize;
  int WSize;
  float Stride;
  int minx, miny, minz, maxx, maxy, maxz;
  char DatasetName[20];
  char RGBFile[200];
  char AlphaFile[200];
  char StfFile[200];
  int HaveShading, HaveDeformation;
  float Dist;
  List AlphaMap, RGBMap;
  RGB **Image;
  unsigned char **ImageMask;
  char ImageFileName[100];
  int EUS;
  float *EUMaxRGB;
  float PrevMaxRGB;
  RGB HIOColor;
  Point PrevHIP, PrevWindowHIP;
  float HIOAlpha;
  Sphere HIO;
  double AverageFrameDuration;
  long FrameCount;

  Renderer(char *dn);
  ~Renderer();
  void setViewPlane(Transform *t);
  void setLightPosition(Point *lpos);
  void render();
  void renderRegion(int x0, int y0, int x1, int y1);
  void renderHIO();
  void renderBox(Box *box);
  void getRGB(float v, RGB *rgb);
  float getAlpha(float v);
  void writeImageFile(char s[]);
  void setHIPPosition(Point *p);
  void setHIPVoxel(int x, int y, int z);
  void setHIPPosition(float x, float y, float z);

  inline float getMin(float a1, float a2, float a3);
  inline float getMax(float a1, float a2, float a3);

  friend void computePixelColor(ExecUnitArgs *eua);
};

#endif
