#include "Sphere.h"

/* ------------------------------------------------------------------------ */
Sphere::Sphere(Point *c, float r)
{
  Center.copy(c);
  Radius = r;
}

/* ------------------------------------------------------------------------ */
Sphere::Sphere()
{
  Radius = 1.0f;
}

/* ------------------------------------------------------------------------ */
Sphere::~Sphere()
{
}

/* ------------------------------------------------------------------------ */
int Sphere::getExtremeIntersections(Ray *ray, float &first, float &last)
{
  float a, b, c;
  
  a = ray->Dx*ray->Dx + ray->Dy*ray->Dy + ray->Dz*ray->Dz;
  b = 2*(ray->Dx*(ray->X0-Center.X) + 
         ray->Dy*(ray->Y0-Center.Y) + 
         ray->Dz*(ray->Z0-Center.Z));
  c = (ray->X0 - Center.X)*(ray->X0 - Center.X)+
    (ray->Y0 - Center.Y)*(ray->Y0 - Center.Y)+
    (ray->Z0 - Center.Z)*(ray->Z0 - Center.Z) - Radius*Radius;

  if(b*b - 4.0f*a*c >= 0.0f)
  {
    if((float)fabs(a) <= 0.0000001f)
    {
      first = -c/b;
      if(first < 0.0f)
        return 0;
      last = first;
      return 1;
    }
    else
    {
      first = (-b-(float)sqrt(b*b-4*a*c))/(2*a);
      last = (-b+(float)sqrt(b*b-4*a*c))/(2*a);
      if(first < 0.0f && last < 0.0f)
        return 0;

      if(first < 0.0f)
        first = 0.00001f;
      return 1;
    }
  }
  return 0;
}

/* ------------------------------------------------------------------------ */
void Sphere::getPointNormal(Point *p, Point *n)
{
  n->copy(p);
  n->sub(&Center);
  n->normalize();
}

/* ------------------------------------------------------------------------ */
void Sphere::print()
{
  printf("Sphere\n");
  printf("\tCenter: ");
  Center.print();
  printf("\tRadius: %f\n", Radius);
}
