#include "Transform.h"

/* ------------------------------------------------------------------------ */
Transform::Transform()
{
  int i, j;
  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
    {
      if(i != j)
        m[i][j] = 0.0f;
      else
        m[i][j] = 1.0f;
    }
}

/* ------------------------------------------------------------------------ */
Transform::Transform(Transform *t)
{
  int i, j;
  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      m[i][j] = t->m[i][j];
}

/* ------------------------------------------------------------------------ */
Transform::Transform(Rotation *r, Point *p)
{
  int i, j;
  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      m[i][j] = r->m[i][j];
  m[0][3] = p->X;
  m[1][3] = p->Y;
  m[2][3] = p->Z;
  m[3][0] = 0.0f;
  m[3][1] = 0.0f;
  m[3][2] = 0.0f;
  m[3][3] = 1.0f;
}

/* ------------------------------------------------------------------------ */
void Transform::copy(Transform *t)
{
  int i, j;
  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      m[i][j] = t->m[i][j];
}

/* ------------------------------------------------------------------------ */
void Transform::copy(Rotation *r, Point *p)
{
  int i, j;
  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      m[i][j] = r->m[i][j];
  m[0][3] = p->X;
  m[1][3] = p->Y;
  m[2][3] = p->Z;
  m[3][0] = 0.0f;
  m[3][1] = 0.0f;
  m[3][2] = 0.0f;
  m[3][3] = 1.0f;
}

/* ------------------------------------------------------------------------ */
void Transform::invert()
{
  float buf[3][3];
  int i, j;
  Rotation r;
  Point p;

  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      buf[i][j] = m[j][i];

  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      m[i][j] = buf[i][j];

  r.copy(buf);
  p.copy(m[0][3], m[1][3], m[2][3]);
  p.rot(&r);
  p.scale(-1.0f);
  m[0][3] = p.X;
  m[1][3] = p.Y;
  m[2][3] = p.Z;
}

/* ------------------------------------------------------------------------ */
void Transform::preMult(Transform *t)
{
  int i, j;
  float aux[4][4];

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      aux[i][j] = t->m[i][0]*m[0][j] + t->m[i][1]*m[1][j] + 
        t->m[i][2]*m[2][j] + t->m[i][3]*m[3][j];

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      m[i][j] = aux[i][j];
}

/* ------------------------------------------------------------------------ */
void Transform::postMult(Transform *t)
{
  int i, j;
  float aux[4][4];

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      aux[i][j] = m[i][0]*t->m[0][j] + m[i][1]*t->m[1][j] + 
        m[i][2]*t->m[2][j] + m[i][3]*t->m[3][j];

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      m[i][j] = aux[i][j];
}

/* ------------------------------------------------------------------------ */
void Transform::preRot(Rotation *r)
{
  int i, j;
  float aux[4][4];

  for(i=0; i<3; i++)
    for(j=0; j<4; j++)
      aux[i][j] = r->m[i][0]*m[0][j] + r->m[i][1]*m[1][j] + 
        r->m[i][2]*m[2][j];

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      m[i][j] = aux[i][j];
}

/* ------------------------------------------------------------------------ */
void Transform::postRot(Rotation *r)
{
  int i, j;
  float aux[4][4];

  for(i=0; i<4; i++)
    for(j=0; j<3; j++)
      aux[i][j] = m[i][0]*r->m[0][j] + m[i][1]*r->m[1][j] + 
        m[i][2]*r->m[2][j];

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      m[i][j] = aux[i][j];
}
