#include "Renderer.h"
#include <sys/types.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h> 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <X11/keysym.h>

int screen;
Window win;
int depth;
Display *display;
GC image_gc;
int *ImageBuffer;
XImage *xim;
XSetWindowAttributes window_attributes;
unsigned long window_mask;
XSizeHints window_hints;
Colormap cm;
float dist = 10;
float angle = 5;
Renderer *R;
char rot_axis='x';
int online = 1;
int frozen = 0;
pthread_mutex_t mtx;
char arg1[100];

/* ------------------------------------------------------------------------ */
Visual* match_visual (Display *dpy, int screen, int depth)
{  
  XVisualInfo tmpl;  
  XVisualInfo *vinfo;
  int nitems;  

  tmpl.screen = screen; 
  tmpl.depth = depth;  
  if (!(vinfo = XGetVisualInfo(dpy, VisualScreenMask | VisualDepthMask, 
                               &tmpl, &nitems)) || nitems <= 0) 
    return (NULL);  

  return (vinfo[0].visual);
}

/* ------------------------------------------------------------------------ */
void createImage()
{
  int x, y, r, g, b;

  for(x=0; x<R->WSize; x++)
    for(y=0; y<R->WSize; y++)
      {
        r = (int) floor(R->Image[y][x].Red*255);
        g = (int) floor(R->Image[y][x].Green*255);
        b = (int) floor(R->Image[y][x].Blue*255);
        ImageBuffer[y*R->WSize+x] = ((b << 16) + (g << 8) + r); 
      }
}

/* ------------------------------------------------------------------------ */
void generateRotX(float alpha, Transform *t)
{
  float a, ca, sa;

  a = alpha/180.0f*3.14159f;
  ca = (float)cos(a);
  sa = (float)sin(a);

  t->m[0][0] = 1.0f;
  t->m[0][1] = 0.0f;
  t->m[0][2] = 0.0f;
  t->m[0][3] = 0.0f;
  t->m[1][0] = 0.0f;
  t->m[1][1] = ca;
  t->m[1][2] = -sa;
  t->m[1][3] = 0.0f;
  t->m[2][0] = 0.0f;
  t->m[2][1] = sa;
  t->m[2][2] = ca;
  t->m[2][3] = 0.0f;
  t->m[3][0] = 0.0f;
  t->m[3][1] = 0.0f;
  t->m[3][2] = 0.0f;;
  t->m[3][3] = 0.0f;
}

/* ------------------------------------------------------------------------ */
void generateRotY(float alpha, Transform *t)
{
  float a, ca, sa;

  a = alpha/180.0f*3.14159f;
  ca = (float)cos(a);
  sa = (float)sin(a);

  t->m[0][0] = ca;
  t->m[0][1] = 0.0f;
  t->m[0][2] = sa;
  t->m[0][3] = 0.0f;
  t->m[1][0] = 0.0f;
  t->m[1][1] = 1.0f;
  t->m[1][2] = 0.0f;
  t->m[1][3] = 0.0f;
  t->m[2][0] = -sa;
  t->m[2][1] = 0.0f;
  t->m[2][2] = ca;
  t->m[2][3] = 0.0f;
  t->m[3][0] = 0.0f;
  t->m[3][1] = 0.0f;
  t->m[3][2] = 0.0f;;
  t->m[3][3] = 0.0f;
}

/* ------------------------------------------------------------------------ */
void generateRotZ(float alpha, Transform *t)
{
  float a, ca, sa;

  a = alpha/180.0f*3.14159f;
  ca = (float)cos(a);
  sa = (float)sin(a);

  t->m[0][0] = ca;
  t->m[0][1] = -sa;
  t->m[0][2] = 0.0f;
  t->m[0][3] = 0.0f;
  t->m[1][0] = sa;
  t->m[1][1] = ca;
  t->m[1][2] = 0.0f;
  t->m[1][3] = 0.0f;
  t->m[2][0] = 0.0f;
  t->m[2][1] = 0.0f;
  t->m[2][2] = 1.0f;
  t->m[2][3] = 0.0f;
  t->m[3][0] = 0.0f;
  t->m[3][1] = 0.0f;
  t->m[3][2] = 0.0f;;
  t->m[3][3] = 0.0f;
}

/* ------------------------------------------------------------------------ */
void rotateVolume()
{
  Transform t, trans;

  while(online)
  {
    switch(rot_axis)
    {
      case 'x':
        generateRotX(angle, &t);
        break;
      case 'y':
        generateRotY(angle, &t);
        break;
      case 'z':
        generateRotZ(angle, &t);
        break;
    };
    trans.copy(&R->T);
    trans.preMult(&t);
    R->setViewPlane(&trans);
    pthread_mutex_lock(&mtx);
    pthread_mutex_unlock(&mtx);
    if(!online)
      break;
    R->render();
#ifdef SAVE_SNAPSHOTS
    char s[50], cmd[100];
    sprintf(s, "../img/%s_%d.pbm", arg1, (int)time(NULL));
    R->writeImageFile(s);
    sprintf(cmd, "cp %s ../img/last.pbm", s);
    system(cmd);
#endif
    createImage();
    XPutImage (display, win, image_gc, xim, 0, 0, 0, 0, 
               R->WSize, R->WSize);
  }
}

/* ------------------------------------------------------------------------ */
void rotateVolumeN()
{
  Transform t, trans;
  FILE *fbr;

  while(R->FrameCount < 11)
  {
    switch(rot_axis)
    {
      case 'x':
        generateRotX(angle, &t);
        break;
      case 'y':
        generateRotY(angle, &t);
        break;
      case 'z':
        generateRotZ(angle, &t);
        break;
    };
    trans.copy(&R->T);
    trans.preMult(&t);
    R->setViewPlane(&trans);
    pthread_mutex_lock(&mtx);
    pthread_mutex_unlock(&mtx);
    if(!online)
      break;
    R->render();
#ifdef SAVE_SNAPSHOTS
    char s[50], cmd[100];
    sprintf(s, "../img/%s_%d.pbm", arg1, (int)time(NULL));
    R->writeImageFile(s);
    sprintf(cmd, "cp %s ../img/last.pbm", s);
    system(cmd);
#endif
    createImage();
    XPutImage (display, win, image_gc, xim, 0, 0, 0, 0, 
               R->WSize, R->WSize);
  }
  fbr = fopen("fbr", "ab");
  fprintf(fbr, "%f\n", R->AverageFrameDuration);
  fclose(fbr);
  XDestroyImage(xim);
  delete R;
  exit(0);
}

/* ------------------------------------------------------------------------ */
void rotateHIO()
{
  Transform t;
  Point p;

  R->setHIPPosition(50.0f, 50.0f, 50.0f);
  while(online)
  {
    switch(rot_axis)
    {
      case 'x':
        generateRotX(angle, &t);
        break;
      case 'y':
        generateRotY(angle, &t);
        break;
      case 'z':
        generateRotZ(angle, &t);
        break;
    };
    p.copy(&R->HIO.Center);
    p.trans(&t);

    R->setHIPPosition(&p);
    pthread_mutex_lock(&mtx);
    pthread_mutex_unlock(&mtx);
    if(!online)
      break;
    R->renderHIO();
#ifdef SAVE_SNAPSHOTS
    char s[50], cmd[100];
    sprintf(s, "../img/%s_%d.pbm", arg1, (int)time(NULL));
    R->writeImageFile(s);
    sprintf(cmd, "cp %s ../img/last.pbm", s);
    system(cmd);
#endif
    createImage();
    XPutImage (display, win, image_gc, xim, 0, 0, 0, 0, 
               R->WSize, R->WSize);
  }
}

/* ------------------------------------------------------------------------ */
void moveHIO()
{
  Point p;

  while(online)
  {
    p.copy(1.0f, 1.0f, 1.0f);
    p.add(&R->DS->avp);
    R->DS->avp.copy(&p);
    pthread_mutex_lock(&mtx);
    pthread_mutex_unlock(&mtx);
    if(!online)
      break;
    R->render();
#ifdef SAVE_SNAPSHOTS
  char s[50], cmd[100];
    sprintf(s, "../img/%s_%d.pbm", arg1, (int)time(NULL));
    R->writeImageFile(s);
    sprintf(cmd, "cp %s ../img/last.pbm", s);
    system(cmd);
#endif
    createImage();
    XPutImage (display, win, image_gc, xim, 0, 0, 0, 0, 
               R->WSize, R->WSize);
  }
}

/* ------------------------------------------------------------------------ */
int main(int argc, char** argv)
{
  char s[50], cmd[100];
  Visual *vis;
  XSizeHints size_hints;
  KeySym key;
  Point *dir, *vup, *last_dir;
  pthread_t pth;
  pthread_attr_t attr;

  if(argc < 2)
  {
    printf("Usage:  vh <dataset base name>\n\n");
    exit(0);
  }

  strcpy(arg1, argv[1]);
  /* Initialize the renderer */
  R = new Renderer(argv[1]);
  last_dir = new Point(&R->INITDIR);
  dir = new Point(last_dir);
  vup = new Point();

  /* Create the window */
  display = XOpenDisplay (NULL);
  screen = DefaultScreen (display);
  depth = 24;
  vis = match_visual(display, screen, depth);

  cm = XCreateColormap(display, RootWindow (display, screen), vis, AllocNone);

  window_attributes.border_pixel = BlackPixel (display, screen);
  window_attributes.background_pixel = BlackPixel (display, screen);
  window_attributes.override_redirect = 0;
  window_attributes.colormap = cm;
  window_mask = CWBackPixel | CWBorderPixel|CWOverrideRedirect|CWColormap;

  win = XCreateWindow (display, RootWindow (display, screen), 0, 0, 
                       R->WSize, R->WSize, 0, depth, CopyFromParent, 
                       vis, window_mask, &window_attributes);

  size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
  size_hints.x = 300;
  size_hints.y = 300;
  size_hints.width = R->WSize;
  size_hints.height = R->WSize;
  size_hints.min_width = R->WSize;
  size_hints.min_height = R->WSize;
  size_hints.max_width = R->WSize;
  size_hints.max_height = R->WSize;

  XSetStandardProperties(display, win, "Volume Renderer" ,NULL,
                         0,NULL,0,&size_hints);

  image_gc = XCreateGC (display, win, 0, 0);
  XMapWindow (display, win);
  XSelectInput(display, win, ExposureMask | KeyPressMask | StructureNotifyMask);

  ImageBuffer = new int[R->WSize*R->WSize];
  xim = XCreateImage (display, CopyFromParent, depth, ZPixmap, 0,
                      (char*)ImageBuffer, R->WSize, R->WSize, 32, R->WSize * 4);

  R->setHIPPosition(100.0f, 100.0f, 100.0f);
  R->setHIPVoxel(0, 0, 0);
  R->DS->avp.copy(1.0f, 1.0f, 1.0f);
  R->render();
#ifdef SAVE_SNAPSHOTS
  sprintf(s, "../img/%s_%d.pbm", arg1, (int)time(NULL));
  R->writeImageFile(s);
  sprintf(cmd, "../cp %s img/last.pbm", s);
  system(cmd);
#endif
  createImage();
  while (1)
  {
	XEvent x_event;
	XNextEvent(display, &x_event);
    switch  (x_event.type) 
    {
      case Expose:
        XPutImage (display, win, image_gc, xim, 0, 0, 0, 0, R->WSize, R->WSize);
        XFlush(display);
        pthread_mutex_init(&mtx, NULL);
        pthread_attr_init(&attr);
        pthread_attr_setschedpolicy(&attr, SCHED_RR);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
        pthread_create(&pth, &attr, (void *(*)(void *))rotateVolume, NULL);
		break;            
      case KeyPress:
        key = XLookupKeysym(&x_event.xkey, 0);
        switch(key)
        {
          case XK_q:
            online = 0;
            pthread_mutex_unlock(&mtx);
            pthread_join(pth, NULL);
            pthread_attr_destroy(&attr);
            pthread_mutex_destroy(&mtx);
            XDestroyImage(xim);
            printf("Average frame duration: %.3f milliseconds\n", 
                   R->AverageFrameDuration/1000);
            delete R;
            delete dir;
            delete last_dir;
            delete vup;
            exit(0);
          case XK_w:
            sprintf(s, "../img/%s_%d.pbm", argv[1], (int)time(NULL));
            R->writeImageFile(s);
            sprintf(cmd, "cp %s ../img/last.pbm", s);
            system(cmd);
            break;
          case XK_f:
            if(!frozen)
            {
              frozen = 1;
              pthread_mutex_lock(&mtx);;
            }
            break;
          case XK_g:
            if(frozen)
            {
              pthread_mutex_unlock(&mtx);;
              frozen = 0;
            }
            break;
          case XK_x:
            rot_axis = 'x';
            break;
          case XK_y:
            rot_axis = 'y';
            break;
          case XK_z:
            rot_axis = 'z';
            break;
          case XK_Up:
            dist += 1.0f;
            printf("The distance step is %.2f.\n", dist);
            break;
          case XK_Down:
            dist -= 1.0f;
            printf("The distance step is %.2f.\n", dist);
            break;
          case XK_Right:
            angle += 1.0f;
            printf("The angle step is %.2f.\n", angle);
            break;
          case XK_Left:
            angle -= 1.0f;
            printf("The angle step is %.2f.\n", angle);
            break;
        };
		break;
    }			
  }
  return 1;
}
